#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <limits.h>

#include <GL/glew.h>
#include <GL/glu.h>

#include <SDL2/SDL.h>
#include <SDL2/SDL_opengl.h>

#include "utils.h"
#include "glutils.h"
#include "mathext.h"
#include "camera.h"
#include "shaders.h"
#include "object3d.h"



#define SCREEN_WIDTH        800
#define SCREEN_HEIGHT       600

#define POSITION            0
#define COLOR               1

#define NEAR  0.1f
#define FAR  100.0f


SDL_Window* g_window = NULL;
SDL_GLContext g_context;

unsigned int shader_program, normals_shader_program;

GLint ProjectionMatrixLocation;
GLint ModelViewMatrixLocation;
GLint NormalMatrixLocation;
GLint LightPositionLocation;
GLint AmbientLightLocation;
GLint TimeLocation;

unsigned int vertex_array;
unsigned int vertex_buffer;


float projection_matrix[16];
float view_matrix[16];
float model_matrix[16];
float modelview_matrix[16];
float normal_matrix[16];
float aux_matrix1[16];
float aux_matrix2[16];

float cam_position[3] = {0.0f, 7.4f, 20.0f};
float world_up[3] = {0.0f, 1.0f, 0.0f};
float cam_front[3] = {0.0f, 0.0f, -1.0f};
t_camera camera;
t_object3d *o;
float light_pos[4] = {0.0f, 0.0f, 3.0f, 0.0f};
float ambient_light[4] = {0.5f, 0.5f, 0.5f, 1.0f};

float delta_time = 0.0f;
float last_frame_time = 0.0f;
int last_mouse_x = INT_MIN;
int last_mouse_y = INT_MIN;

int wireframe = 0;
int show_depth_map = 0;


float points[] = {
    -0.5f,  0.5f, 1.0f, 0.0f, 0.0f, // top-left
     0.5f,  0.5f, 0.0f, 1.0f, 0.0f, // top-right
     0.5f, -0.5f, 0.0f, 0.0f, 1.0f, // bottom-right
    -0.5f, -0.5f, 1.0f, 1.0f, 0.0f  // bottom-left
};

int init();
int initGL();
int set_shaders();
int init_buffers();
void cleanup();
void update();
void render();




int init() {
    // Initialize SDL
    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
        printf("Błąd podczas inicjalizacji silnika SDL2: %s\n", SDL_GetError());
        return 0;
    }

    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 1);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);

    g_window = SDL_CreateWindow("normal vectors", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
                                SCREEN_WIDTH, SCREEN_HEIGHT, SDL_WINDOW_OPENGL | SDL_WINDOW_SHOWN);
    if (g_window == NULL) {
        printf("Błąd podczas tworzenia okna: %s\n", SDL_GetError());
        return 0;
    }

    g_context = SDL_GL_CreateContext(g_window);
    if (g_context == NULL) {
        printf("Błąd podczas tworzenia kontekstu OpenGL: %s\n", SDL_GetError());
        return 0;
    }

    if (SDL_GL_SetSwapInterval(1) < 0) {
        printf( "Błąd podczas ustawiania VSync: %s\n", SDL_GetError());
    }

    init_camera(&camera, cam_position, cam_front, world_up, -90.0f, 0.0f,
                CAMERA_DEFAULT_SPEED_FACTOR, CAMERA_DEFAULT_MOUSE_SENSITIVITY, 45.0f);
    if (!initGL()) {
        printf("Błąd podczas inicjalizacji OpenGL!\n");
        return 0;
    }

    //printf("VENDOR: %s\n", glGetString(GL_VENDOR));
    //printf("RENDERER: %s\n", glGetString(GL_RENDERER));
    //printf("VERSION: %s\n", glGetString(GL_VERSION));
    //printf("SHADING_LANG: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));
    //printf("EXTENSIONS: %s\n", glGetString(GL_EXTENSIONS));

    last_frame_time = SDL_GetTicks();
    return 1;
}

int initGL() {
    GLenum err = GL_NO_ERROR;
    unsigned int width = SCREEN_WIDTH;
    unsigned int height = SCREEN_HEIGHT;
    int i;

    glewExperimental = GL_TRUE;
    err = glewInit();
    while (check_gl_error() == GL_INVALID_ENUM) {};
    if (err != GLEW_OK) {
        printf("glewInit failed: %s\n", glewGetErrorString(err));
        return 0;
    }

    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

    glViewport(0, 0, width, height);

    if (!set_shaders()) {
        printf("set_shaders failed\n");
        return 0;
    }

    if (!init_buffers()) {
        printf("init_buffers failed\n");
        return 0;
    }


    glEnable(GL_DEPTH_TEST);
    //glDepthFunc(GL_ALWAYS);
    glDepthFunc(GL_LEQUAL);

    float aspect = (float)SCREEN_WIDTH / (float)SCREEN_HEIGHT;
    get_perspective_matrix(projection_matrix, camera.fov, aspect, NEAR, FAR);

    o = object3d_load_from_obj_file("nanosuit/nanosuit.obj", 0);
    if (!o) {
        fprintf(stderr, "object loading failed\n");
        return 0;
    }

    return 1;
}


int set_shaders() {
    shader_program = glCreateProgram();
    glAttachShader(shader_program, load_shader(GL_VERTEX_SHADER, "vertex_shader.glsl"));
    glAttachShader(shader_program, load_shader(GL_FRAGMENT_SHADER, "fragment_shader.glsl"));
    glAttachShader(shader_program, load_shader(GL_GEOMETRY_SHADER, "geometry_shader.glsl"));
    link_and_validate_program(shader_program);

    normals_shader_program = glCreateProgram();
    glAttachShader(normals_shader_program, load_shader(GL_VERTEX_SHADER, "vertex_shader_normals.glsl"));
    glAttachShader(normals_shader_program, load_shader(GL_FRAGMENT_SHADER, "fragment_shader_normals.glsl"));
    glAttachShader(normals_shader_program, load_shader(GL_GEOMETRY_SHADER, "geometry_shader_normals.glsl"));
    link_and_validate_program(normals_shader_program);

    glUseProgram(0);

    return 1;
}


int init_buffers() {
    glGenVertexArrays(1, &vertex_array);
    if (!vertex_array) {
        return 0;
    }
    glBindVertexArray(vertex_array);
    glGenBuffers(1, &vertex_buffer);
    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer);
    glBufferData(GL_ARRAY_BUFFER, sizeof(points), points, GL_STATIC_DRAW);
    glVertexAttribPointer(POSITION, 2, GL_FLOAT, GL_FALSE, 5*sizeof(float), (void*)0);
    glVertexAttribPointer(COLOR, 3, GL_FLOAT, GL_FALSE, 5*sizeof(float), (void*)(2*sizeof(float)));
    glEnableVertexAttribArray(POSITION);
    glEnableVertexAttribArray(COLOR);

    glBindVertexArray(0);

    return 1;
}


void cleanup() {
    if (o) {
        object3d_free(o);
    }

    if (shader_program) {
        glDeleteProgram(shader_program);
    }
    if (normals_shader_program) {
        glDeleteProgram(normals_shader_program);
    }


    if (vertex_buffer) {
        glDeleteBuffers(1, &vertex_buffer);
    }
    if (vertex_array) {
        glDeleteVertexArrays(1, &vertex_array);
    }


    if (g_context) {
        SDL_GL_DeleteContext(g_context);
        g_context = NULL;
    }

    if (g_window) {
        SDL_DestroyWindow(g_window);
        g_window = NULL;
    }

    SDL_Quit();
}

void update() {
    update_camera_vectors(&camera);
    get_camera_look_at_matrix(view_matrix, &camera);

    light_pos[0] = camera.front[0];
    light_pos[1] = camera.front[1];
    light_pos[2] = camera.front[2];
}

void render() {
    char buf[255];

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    if (wireframe) {
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    } else {
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }

    get_identity_matrix(model_matrix);
    mult_matrices4x4(modelview_matrix, view_matrix, model_matrix);
    calc_normal_matrix(normal_matrix, modelview_matrix);

    glUseProgram(shader_program);

    ProjectionMatrixLocation = glGetUniformLocation(shader_program, "ProjectionMatrix");
    ModelViewMatrixLocation  = glGetUniformLocation(shader_program, "ModelViewMatrix");
    NormalMatrixLocation     = glGetUniformLocation(shader_program, "NormalMatrix");
    LightPositionLocation    = glGetUniformLocation(shader_program, "LightPosition");
    AmbientLightLocation     = glGetUniformLocation(shader_program, "AmbientLight");

    object3d_set_vertex_locations(o, glGetAttribLocation(shader_program, "vTangent"),
                        glGetAttribLocation(shader_program, "vNormal"),
                        glGetAttribLocation(shader_program, "vTexCoord"),
                        glGetAttribLocation(shader_program, "vPosition"));

    object3d_set_property_locations(o, OBJ_KN, -1, glGetUniformLocation(shader_program, "NormalTexture"), -1);
    object3d_set_property_locations(o, OBJ_KD, -1, glGetUniformLocation(shader_program, "DiffuseTexture"), -1);
    object3d_set_property_locations(o, OBJ_KS, -1, glGetUniformLocation(shader_program, "SpecularTexture"), -1);

    glUniformMatrix4fv(ProjectionMatrixLocation, 1, GL_TRUE, projection_matrix);
    glUniformMatrix4fv(ModelViewMatrixLocation,  1, GL_TRUE, modelview_matrix);
    glUniformMatrix3fv(NormalMatrixLocation,     1, GL_TRUE, normal_matrix);
    glUniform4fv      (LightPositionLocation,    1, light_pos);
    glUniform4fv      (AmbientLightLocation,     1, ambient_light);

    object3d_render(o);
    glUseProgram(0);




    glUseProgram(normals_shader_program);

    ProjectionMatrixLocation = glGetUniformLocation(normals_shader_program, "ProjectionMatrix");
    ModelViewMatrixLocation  = glGetUniformLocation(normals_shader_program, "ModelViewMatrix");
    NormalMatrixLocation     = glGetUniformLocation(normals_shader_program, "NormalMatrix");
    LightPositionLocation    = glGetUniformLocation(normals_shader_program, "LightPosition");
    AmbientLightLocation     = glGetUniformLocation(normals_shader_program, "AmbientLight");

    object3d_set_vertex_locations(o, glGetAttribLocation(normals_shader_program, "vTangent"),
                        glGetAttribLocation(normals_shader_program, "vNormal"),
                        glGetAttribLocation(normals_shader_program, "vTexCoord"),
                        glGetAttribLocation(normals_shader_program, "vPosition"));

    object3d_set_property_locations(o, OBJ_KN, -1, glGetUniformLocation(normals_shader_program, "NormalTexture"), -1);
    object3d_set_property_locations(o, OBJ_KD, -1, glGetUniformLocation(normals_shader_program, "DiffuseTexture"), -1);
    object3d_set_property_locations(o, OBJ_KS, -1, glGetUniformLocation(normals_shader_program, "SpecularTexture"), -1);


    glUniformMatrix4fv(ProjectionMatrixLocation, 1, GL_TRUE, projection_matrix);
    glUniformMatrix4fv(ModelViewMatrixLocation,  1, GL_TRUE, modelview_matrix);
    glUniformMatrix3fv(NormalMatrixLocation,     1, GL_TRUE, normal_matrix);
    glUniform4fv      (LightPositionLocation,    1, light_pos);
    glUniform4fv      (AmbientLightLocation,     1, ambient_light);

    object3d_render(o);
    glUseProgram(0);

    //draw_text(0.0f, 0.0f, char *str, float size, float r, float g, float b, float alpha);
}



int main(int argc, char* args[]) {
    int quit;
    SDL_Event e;
    int nrAttributes;
    float current_frame_time;

    if (!init()) {
        printf("Błąd inicjalizacji!\n");
        cleanup();
        return 1;
    }

    quit = 0;
    while (!quit) {
        current_frame_time = SDL_GetTicks();
        delta_time = current_frame_time - last_frame_time;
        last_frame_time = current_frame_time;

        while (SDL_PollEvent(&e) != 0) {
            if (e.type == SDL_QUIT) {
                quit = 1;
                break;
             } else if (e.type == SDL_KEYDOWN) {
                switch (e.key.keysym.sym) {
                    case SDLK_UP:
                        camera_move(&camera, CAMERA_MOVEMENT_FORWARD, delta_time);
                        break;
                    case SDLK_DOWN:
                        camera_move(&camera, CAMERA_MOVEMENT_BACKWARD, delta_time);
                        break;
                    case SDLK_LEFT:
                        camera_move(&camera, CAMERA_MOVEMENT_LEFT, delta_time);
                        break;
                    case SDLK_RIGHT:
                        camera_move(&camera, CAMERA_MOVEMENT_RIGHT, delta_time);
                        break;
                    case SDLK_w:
                        wireframe = 1 - wireframe;
                        break;
                    case SDLK_d:
                        show_depth_map = 1 - show_depth_map;
                        break;
                    case SDLK_q:
                        quit = 1;
                        break;
                    case SDLK_PAGEUP:
                        camera_move(&camera, CAMERA_MOVEMENT_UP, delta_time);
                        break;
                    case SDLK_PAGEDOWN:
                        camera_move(&camera, CAMERA_MOVEMENT_DOWN, delta_time);
                        break;
                }

            } else if (e.type == SDL_MOUSEMOTION) {
                int x, y;
                float x_offset, y_offset;

                SDL_GetMouseState(&x, &y);

                x_offset = x - last_mouse_x;
                y_offset = last_mouse_y - y;

                if (last_mouse_x != INT_MIN && last_mouse_y != INT_MIN && fabs(x_offset) < 100.0f && fabs(y_offset) < 100.0f) {
                    if (SDL_GetModState() & KMOD_SHIFT == KMOD_SHIFT) {
                        camera_rotate(&camera, x_offset, y_offset, 1);
                    }
                }

                last_mouse_x = x;
                last_mouse_y = y;

            } else if (e.type == SDL_MOUSEWHEEL) {
                camera_zoom(&camera, e.wheel.y);

                float aspect = (float)SCREEN_WIDTH / (float)SCREEN_HEIGHT;
                get_perspective_matrix(projection_matrix, camera.fov, aspect, NEAR, FAR);
            }
        }
        update();
        render();

        SDL_GL_SwapWindow(g_window);
    }

    cleanup();

    return 0;
}
